# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Philippe Biondi <phil@secdev.org>

"""
Package of extension modules that have to be loaded explicitly.
"""

# Make sure config is loaded
import scapy.config  # noqa: F401
